from typing import Optional, OrderedDict, Dict, Any, Union, Literal, NamedTuple, Tuple, Callable
from dataclasses import dataclass, field

import time
import logging
import urllib
import urllib3
import json
import io

import requests
import requests_toolbelt

from .baseclient import BaseClient
from .requestutils import BandwidthUsage, try_request, get_json
from .usercookie import UserCookie
from .options import RequestOptions, LoginPolicy, LuweiCookieFormat
from .objects import Board, ThreadPage, BoardThread
from .utils import current_timestamp_ms_offset_to_utc8
from .exceptions import ShouldNotReachException, RequiresLoginException, NoPermissionException, ResourceNotExistsException, GatekeptException


@dataclass
class Client(BaseClient):
    """
        实现各类基础操作的客户端类。
    """

    def get_board_page(self, board_id: int, page: int, options: RequestOptions = {}) -> Tuple[Board, BandwidthUsage]:
        """
        获取指定板块的指定页。

        Parameters
        ----------
        board_id : int
            版块 ID。
        page : int
            页数。
        options : RequestOptions
            请求选项。
        """

        needs_login = self.board_page_requires_login(
            page=page, options=options)
        if needs_login and not self.has_cookie(options):
            raise RequiresLoginException()

        logging.debug(f"将获取版块：{board_id} 第 {page} 页，将会登录：{needs_login}")

        def request_fn():
            threads, bandwidth_usage = self._get_json(
                path=f'/Api/showf', options=options, needs_login=needs_login,
                id=board_id,
                page=page,
            )
            return list(map(lambda thread: BoardThread(thread), threads)), bandwidth_usage

        (board_page, bandwidth_usage) = try_request(
            request_fn, f"获取版块 {id} 第 {page} 页", self.get_max_attempts(options))

        return board_page, bandwidth_usage

    def get_thread_page(self, id: int, page: int, options: RequestOptions = {}, for_analysis: bool = False) -> Tuple[ThreadPage, BandwidthUsage]:
        """
        获取指定串的指定页。

        Parameters
        ----------
        id : int
            串号。
        page : int
            页数。
        options : RequestOptions
            请求选项。
        for_analysis : bool
            如果为真，将会过滤掉与分析无关的内容，以方便分析。
        """

        needs_login = self.thread_page_requires_login(
            page=page, options=options)
        if needs_login and not self.has_cookie(options):
            raise RequiresLoginException()

        logging.debug(f"将获取串：{id} 第 {page} 页，将会登录：{needs_login}")

        def request_fn():
            thread_page_json, bandwidth_usage = self._get_json(
                path=f'/Api/thread/id/{id}', options=options, needs_login=needs_login,
                page=page,
            )
            if thread_page_json == '该主题不存在':
                raise ResourceNotExistsException()

            return ThreadPage(thread_page_json), bandwidth_usage

        (thread_page, bandwidth_usage) = try_request(
            request_fn, f"获取串 {id} 第 {page} 页", self.get_max_attempts(options))

        if for_analysis:
            thread_page.replies = list(filter(
                lambda post: post.user_id != "芦苇", thread_page.replies))

        return thread_page, bandwidth_usage

    def _get_json(self, path: str, options: RequestOptions, needs_login: bool = False, **queries) -> Tuple[OrderedDict, BandwidthUsage]:
        session = self._make_session(options=options, needs_login=needs_login)
        url = self._make_request_url(path=path, **queries)
        return get_json(session, url)

    def _make_request_url(self, path: str, **queries) -> str:
        queries = OrderedDict(queries)

        # 添加通用参数
        if self.appid != None:
            queries["appid"] = self.appid
        queries["__t"] = current_timestamp_ms_offset_to_utc8()

        base_url = f'https://{self.host}{path}'
        return base_url + '?' + urllib.parse.urlencode(queries)

    def page_requires_login(self, page: int, gate_keeper: int, options: RequestOptions = {}) -> bool:
        """
        判断页面是否需要登录才能正常阅读。

        Parameters
        ----------
        page : int
            要请求的页面页数。
        gate_keeper : int
            从下页起会出现「卡页」现象的页数，
            即如果不登录，访问往后的页面都会响应该页的内容。

            2021-03-07: 目前无论是串还是板块，都是从第101页起重复第100页的内容。
        options : RequestOptions
            请求设置。

        Returns
        -------
        是否需要登录。
        """

        login_policy = self.get_login_policy(options)
        has_cookie = self.has_cookie(options)

        if login_policy == "enforce":
            return True

        if login_policy == "when_has_cookie":
            return has_cookie or page > gate_keeper
        elif login_policy in ("always_no", "when_required"):
            return page > gate_keeper

        raise ShouldNotReachException()

    def thread_page_requires_login(self, page: int, options: RequestOptions = {}) -> bool:
        return self.page_requires_login(
            page=page,
            gate_keeper=self.get_thread_gatekeeper_page_number(options),
            options=options,
        )

    def board_page_requires_login(self, page: int, options: RequestOptions = {}) -> bool:
        gk_pn = self.get_board_gatekeeper_page_number(options)
        if page > gk_pn:  # TODO: 放在这里是不是不太合适？
            raise GatekeptException(
                context='check_if_board_page_requires_login',
                current_page_number=None,
                gatekeeper_post_id=None,
            )
        return self.page_requires_login(
            page=page,
            gate_keeper=gk_pn,
            options=options,
        )

    def get_thread_gatekeeper_page_number(self, options: RequestOptions = {}) -> int:
        return self._get_option_value(options, "thread_gatekeeper_page_number", 100)

    def get_board_gatekeeper_page_number(self, options: RequestOptions = {}) -> int:
        return self._get_option_value(options, "board_gatekeeper_page_number", 100)



from typing import NamedTuple, Callable, Any, OrderedDict

import logging
import urllib3
import json
import io

import requests
import requests_toolbelt

from .exceptions import ResourceNotExistsException


class BandwidthUsage(NamedTuple):
    """
    用于封装与客户端操作产生的流量有关的信息。
    """
    uploaded: int
    """上传字节数。"""
    downloaded: int
    """下载字节数。"""


def try_request(fn: Callable[[], Any], description: str, max_attempts: int) -> Any:
    """
    尝试进行请求。

    Parameters
    ----------
    fn : Callable[[], Any]
        请求本身。
    description : str
        对于请求的可读描述。
    max_attempts : int
        最多可尝试的次数。
    """

    for i in range(1, max_attempts + 1):
        try:
            return fn()
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
            # 对于连接、超时等可以重试的问题进行重试
            msg = f'执行「{description}」失败：{e}。'
            if i < max_attempts:
                logging.warning(msg + f'将会重试。尝试次数：{i}/{max_attempts}')
            else:
                logging.error(msg + f'已经失败 {max_attempts} 次，超过最大尝试次数，放弃')
        except Exception as e:
            msg = f'执行「{description}」失败：{e}。将不重试，放弃'
            if isinstance(e, requests.exceptions.HTTPError):
                # pylint: disable=maybe-no-member
                if e.response != None and e.response.status_code == 404:
                    # 如果能够确认是 404，那就抛出 :class:`ResourceNotExistsException`
                    e = ResourceNotExistsException()
            elif isinstance(e, requests.exceptions.RequestException):
                # pylint: disable=maybe-no-member
                dump = requests_toolbelt.utils.dump.dump_all(e.response)
                msg += "。dump：" + dump.decode('utf-8')

            logging.error(msg)

            raise e


def get_json(session: requests.Session, url: str):
    # TODO: 不要 hardcode timeout
    with session.get(url, stream=True, timeout=20) as resp:
        resp.raise_for_status()
        raw_content = resp.raw.read()

    bandwidth_usage = BandwidthUsage(
        __calculate_request_size(resp),
        __calculate_response_size(resp, raw_content),
    )

    headers = {
        'Content-Encoding': resp.headers.get('content-encoding', '')
    }

    with io.BytesIO(raw_content) as f:
        fake_resp = urllib3.response.HTTPResponse(
            body=f,
            headers=headers,
        )
        decoded_content = fake_resp.data
        obj = json.loads(decoded_content, object_pairs_hook=OrderedDict)

    return obj, bandwidth_usage


def __calculate_request_size(resp: requests.Response) -> BandwidthUsage:
    """
    …

    See: https://stackoverflow.com/a/33217154
    """

    request_line_size = len(resp.request.method) + \
        len(resp.request.path_url) + 12
    request_size = request_line_size + \
        __calculate_header_size(resp.request.headers) + \
        int(resp.request.headers.get("content-length", 0)
            )  # 没有 body 就不会生成 Content-Length?

    return request_size


def __calculate_response_size(resp: requests.Response, raw_content):

    response_line_size = len(resp.reason) + 15
    response_size = response_line_size + \
        __calculate_header_size(resp.headers) + \
        len(raw_content)

    return response_size


def __calculate_header_size(headers) -> int:
    return sum(len(key) + len(value) + 4 for key, value in headers.items()) + 2



from typing import Any, Dict, Optional, Set
from dataclasses import dataclass, field

from datetime import datetime

import anobbsclient

from .walktarget import WalkTargetInterface


@dataclass
class BoardWalkTargetState:
    stop_before_datetime: datetime
    latest_seen_datetime: Optional[datetime] = None

    seen_thread_ids: Set = field(default_factory=set)

    found_last_replies_before_stop_datetime: bool = False
    should_back_to_begin: bool = False


@dataclass(frozen=True)
class BoardWalkTarget(WalkTargetInterface):

    board_id: int
    """要遍历的版块的id。"""

    # overriding
    start_page_number: int

    stop_before_datetime: datetime
    """
    停止于时间。
    在看到最后回复时间早于此时间时停止。
    """

    def create_state(self) -> BoardWalkTargetState:
        return BoardWalkTargetState(stop_before_datetime=self.stop_before_datetime)

    # overriding
    def get_page(self, current_page_number: int,
                 client: anobbsclient.Client, options: anobbsclient.RequestOptions
                 ) -> [anobbsclient.Board, anobbsclient.BandwidthUsage]:
        return client.get_board_page(
            board_id=self.board_id, page=current_page_number,
            options=options,
        )

    # overriding
    def check_gatekept(self, current_page_number: int,
                       current_page: anobbsclient.Board,
                       client: anobbsclient.Client, options: anobbsclient.RequestOptions,
                       g: BoardWalkTargetState):

        if current_page_number == 1:
            # 在第一页记录这一轮见过的最晚的时间
            g.latest_seen_datetime = current_page[0].last_modified_time

        stop_before_datetime = g.stop_before_datetime
        if current_page[-1].last_modified_time < stop_before_datetime:
            # 如果该页最后一串超过停止时间
            g.found_last_replies_before_stop_datetime = True
            for (i, thread) in enumerate(current_page):
                if thread.last_modified_time < stop_before_datetime:
                    current_page[:] = current_page[:i]

        # 借用这里去重
        current_page[:] = [
            thread for thread in current_page if thread.id not in g.seen_thread_ids]
        for thread in current_page:
            g.seen_thread_ids.add(thread.id)

        # TODO: 超过100页无论是否登录都会卡页，应该给下面的方法改个名
        if client.board_page_requires_login(current_page_number):
            raise anobbsclient.GatekeptException(
                context='board_page_number',
                current_page_number=current_page_number,
                gatekeeper_post_id=None,
            )

    # overriding
    def should_stop(self, current_page: anobbsclient.Board, current_page_number: int,
                    client: anobbsclient.Client, options: anobbsclient.RequestOptions,
                    g: BoardWalkTargetState) -> bool:

        if g.found_last_replies_before_stop_datetime:
            if current_page_number == 1:
                return True
            # 如果不本身就在第1页则回到第1页，防止遗漏期间被顶上去的串
            g.should_back_to_begin = True
            g.stop_before_datetime = g.latest_seen_datetime
            g.latest_seen_datetime = None
        return False

    # overriding
    def get_next_page_number(self, current_page_number: int, g: BoardWalkTargetState):
        # 为了防止遗漏遍历途中被顶上去的串，如果遍历不在第1页便停止，则回到第1页
        if g.should_back_to_begin:
            g.should_back_to_begin = False
            return 1

        return current_page_number + 1



from typing import Optional, TypedDict, Union, Literal

from .usercookie import UserCookie

LoginPolicy = Union[
    Literal["enforce"],
    Literal["when_has_cookie"],
    Literal["when_required"],
    Literal["always_no"],
]
"""
是否向服务端告知登录的策略。

Cases
-----
"enforce"
    无视其他条件，必须设置饼干，并使用该饼干请求服务器。
    如果没有设置用户饼干，将抛出 :exc:`RequiresLoginException`。

"when_has_cookie"
    只要设置了用户饼干，就使用该饼干请求服务器。

"when_required"
    只有在需要时（如获取串的 100 页之后的页面）才使用设置的饼干，否则会以无饼干状态请求服务器。

"always_no"
    即使设置了饼干，也会以无饼干状态请求服务器。

"""


class LuweiCookieFormat(TypedDict, total=False):
    expires: str


class RequestOptions(TypedDict, total=False):
    """
    客户端相关的请求设置。
    """

    user_cookie: UserCookie
    """
    使用的用户饼干，默认为 ``None``。

    若为 ``None``, 则不使用饼干。
    """

    login_policy: LoginPolicy
    """饼干登录的策略。"""

    board_gatekeeper_page_number: int
    """版块「卡99」的页数，默认为 ``100``。"""
    thread_gatekeeper_page_number: int
    """
    串「卡99」的页数，默认为 ``100``。

    如果操作获取串的页数超过该值，但并无可用饼干，会抛出 :exc:`RequiresLoginException`。
    """

    uses_luwei_cookie_format: Union[Literal[False], LuweiCookieFormat]
    """
    是否使用芦苇客户端的格式填充 HTTP ``Cookie`` Header，默认为 ``False``。

    芦苇客户端填充的 ``Cookie`` Header 包含不必要的 ``expires``, ``domains`` 及 ``path`` 字段，大概率是填错了。  
    为了保持与芦苇客户端一致，可以开启此项。
    """

    max_attempts: int
    """最多由于网络连接问题进行尝试的次数，默认为 ``5``。"""



from typing import Tuple, Dict, Any
from dataclasses import dataclass
import abc

import anobbsclient


@dataclass(frozen=True)
class WalkTargetInterface:

    start_page_number: int
    """遍历开始的页数。"""

    @abc.abstractmethod
    def create_state(self) -> Any:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_page(self, current_page_number: int,
                 client: anobbsclient.Client, options: anobbsclient.RequestOptions
                 ) -> Tuple[Any, anobbsclient.BandwidthUsage]:
        raise NotImplementedError()

    @abc.abstractmethod
    def check_gatekept(self, current_page_number: int,
                       current_page: Any,
                       client: anobbsclient.Client, options: anobbsclient.RequestOptions,
                       g: Dict[str, Any]):
        raise NotImplementedError()

    @abc.abstractmethod
    def should_stop(self, current_page: Any, current_page_number: int,
                    client: anobbsclient.Client, options: anobbsclient.RequestOptions,
                    g: Dict[str, Any]) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_next_page_number(self, current_page_number: int, g: Dict[str, Any]):
        raise NotImplementedError()



from typing import Optional
from dataclasses import dataclass, field


@dataclass
class ShouldNotReachException(Exception):
    """
    到达了不应该到达的代码时会抛出的异常。
    """
    pass


@dataclass
class ClientException(Exception):
    """
    与客户端有关的异常。
    """

    message: str
    """记载可读的异常信息。"""


@dataclass
class NoPermissionException(ClientException):
    """
    进行没有权限的操作时会抛出的异常。
    """

    def __init__(self, extra_message: Optional[str] = None):
        message = "无操作权限"
        if extra_message != None:
            message += "：" + extra_message
        super(NoPermissionException, self).__init__(
            message=message,
        )


@dataclass
class RequiresLoginException(NoPermissionException):
    """
    需要登录而未登录时会抛出的异常。
    """

    def __init__(self):
        super(RequiresLoginException, self).__init__(
            extra_message="操作需要有效饼干，但客户端并未设置饼干",
        )


@dataclass
class GatekeptException(NoPermissionException):
    """
    发现出现「卡99」现象时会抛出的异常。
    """

    context: str
    current_page_number: Optional[int]
    gatekeeper_post_id: Optional[int] = field(default=None)
    """
    对于版块，此处为空。

    TODO: 依版块/串细分
    """

    def __init__(self, context: str, current_page_number: int, gatekeeper_post_id: int):
        super(GatekeptException, self).__init__(
            extra_message="出现「卡99现象」",
        )

        self.context = context
        self.current_page_number = current_page_number
        self.gatekeeper_post_id = gatekeeper_post_id


@dataclass
class ResourceNotExistsException(ClientException):
    def __init__(self):
        super(ResourceNotExistsException, self).__init__(
            message="目标资源不存在",
        )


@dataclass
class UnreachableLowerBoundPostIDException(ClientException):
    """
    在从后向前遍历时串，发生「到达串的头部时还没有到达下界」这一诡异现象时抛出的异常？
    """

    lower_bound_post_id: int

    def __init__(self, lower_bound_post_id: int):
        super(UnreachableLowerBoundPostIDException, self).__init__(
            message="串号下界比串首串号小",
        )

        self.lower_bound_post_id = lower_bound_post_id


@dataclass
class UnexpectedLowerBoundPostIDException(ClientException):

    current_page_number: int
    expected_lower_bound_page_number: int
    lower_bound_post_id: int

    def __init__(self, current_page_number: int, expected_lower_bound_page_number: int, lower_bound_post_id: int):
        super(UnexpectedLowerBoundPostIDException, self).__init__(
            message="非预期地提前遇到了串号下界",
        )

        self.current_page_number = current_page_number
        self.expected_lower_bound_page_number = expected_lower_bound_page_number
        self.lower_bound_post_id = lower_bound_post_id



from typing import Optional, Dict, Any, Union, Literal, NamedTuple
from dataclasses import dataclass, field

import requests
from http.cookiejar import CookieJar

from .options import RequestOptions, UserCookie, LoginPolicy, LuweiCookieFormat
from .exceptions import RequiresLoginException


@dataclass
class BaseClient:
    """
    客户端基类。

    不实现实际的请求功能。
    """

    user_agent: str
    """发送请求时要使用的 User Agent。"""

    host: str
    """要请求的 API 服务器的主机名。"""

    appid: Optional[str] = None
    """appid。"""

    default_request_options: RequestOptions = field(default_factory=dict)
    """默认的发送请求时的相关设置。"""

    cookiejar_store: Dict[str, CookieJar] = field(default_factory=dict)
    """
    每个饼干配有独立的 :class:`CookieJar`。
    键为 ``userhash``。
    
    这是因为服务器响应可能会根据饼干要求添加不同的新 cookies，
    这么做可以防止不同饼干间新 cookies 的混淆。

    FIXME: 线程不安全。
    """

    def _make_session(self, options: RequestOptions, needs_login: bool = False) -> requests.Session:
        """
        根据请求设置创建一个新的会话。

        Parameters
        ----------
        options : RequestOptions
            请求设置。
        needs_login : bool
            是否需要携带饼干。

        Returns
        -------
        新的会话。
        """

        session = requests.Session()
        self.__setup_headers(session, options=options,
                             needs_login=needs_login)
        return session

    def __setup_headers(self, session, options: RequestOptions, needs_login: bool = False):
        """
        根据请求选项设置好会话的 headers。

        Parameters
        ----------
        session : requests.Session
            会话。
        options : RequestOptions
            请求设置。
        needs_login : bool
            是否需要登录。
        """

        if needs_login:
            # 若需要登录，配置 cookies

            user_cookie = self.get_user_cookie(options)
            if user_cookie is None:
                raise RequiresLoginException()

            if user_cookie.userhash in self.cookiejar_store:
                cookiejar = self.cookiejar_store[user_cookie.userhash]
            else:
                cookiejar: CookieJar = requests.cookies.cookiejar_from_dict({})
                cookie = requests.cookies.create_cookie(
                    name="userhash", value=user_cookie.userhash, domain=self.host,
                )
                cookiejar.set_cookie(cookie)
                self.cookiejar_store[user_cookie.userhash] = cookiejar
            session.cookies = cookiejar

        luwei_cookie_format = self.get_uses_luwei_cookie_format(options)
        if isinstance(luwei_cookie_format, dict):
            # 芦苇岛似乎搞错了要放置的 cookies。
            # 如有要求，便会照着芦苇岛将错就错
            for (k, v) in {
                "expires": luwei_cookie_format["expires"],
                "domains": self.host,
                "path": "/",
            }.items():
                if k not in session.cookies.keys():
                    cookie = requests.cookies.create_cookie(
                        name=k, value=v, domain=self.host,
                    )
                    session.cookies.set_cookie(cookie)

        session.headers.update({
            "Accept": "application/json",
            "User-Agent": self.user_agent,
            "Accept-Language": "en-us",
            "Accept-Encoding": "gzip, deflate, br",
        })

    def _get_option_value(self, external_options: RequestOptions, key: str, default: Any = None) -> Any:
        """
        获取请求设置中指定键的值。

        如果外部请求设置中有值，优先返回来自外部的值；
        其次可能的话返回内部默认请求设置中的值；
        否则返回指定的默认值。

        Parameters
        ----------
        external_options : RequestOptions
            外部请求的设置。
        key : str
            要获取其值的键。
        default : Any
            内外都没有值时的 fallback。

        Returns
        -------
        要获取的值。
        """
        return (
            external_options.get(key, None)
            or self.default_request_options.get(key, default)
        )

    def get_user_cookie(self, options: RequestOptions = {}) -> UserCookie:
        """获取用户饼干。"""
        if self.get_login_policy(options) == "always_no":
            return None
        return self._get_option_value(options, "user_cookie")

    def has_cookie(self, options: RequestOptions = {}) -> bool:
        """返回是否设置了用户饼干。"""
        return self.get_user_cookie(options) != None

    def get_login_policy(self, options: RequestOptions = {}) -> LoginPolicy:
        """获取登录策略，见 :class:`LoginPolicy`。"""
        return self._get_option_value(options, "login_policy", "when_required")

    def get_uses_luwei_cookie_format(self, options: RequestOptions = {}) -> Union[Literal[False], LuweiCookieFormat]:
        """获取是否要照着芦苇岛岛规则设置请求的 cookies。"""
        return self._get_option_value(options, "uses_luwei_cookie_format", False)

    def get_max_attempts(self, options: RequestOptions = {}) -> int:
        """
        获取最大尝试次数。

        默认为5。
        """
        return self._get_option_value(options, "max_attempts", 5)



from typing import Optional
from dataclasses import dataclass


@dataclass(frozen=True)
class UserCookie:
    """
    代表一个用户饼干。
    """

    userhash: str
    """饼干的 ``userhash``。"""

    mark_name: Optional[str] = None
    """辅助标记用的饼干名，可不填。"""



import anobbsclient

from .walktarget import WalkTargetInterface


def create_walker(target: WalkTargetInterface, client: anobbsclient.Client, options: anobbsclient.RequestOptions = None):

    g = target.create_state()
    if options is None:
        options = {}
    current_pn = target.start_page_number
    while True:
        # 获取页面
        (current_page, usage) = target.get_page(current_pn, client, options)
        # 检查是否卡页（卡页则抛异常）
        target.check_gatekept(current_pn, current_page, client, options, g)
        # TODO: 分离出 preprocess 进行如剪裁回复之类的操作？
        # 检查是否满足终止条件
        should_stop = target.should_stop(
            current_page, current_pn, client, options, g)

        # 产出当前页
        yield (current_pn, current_page, usage)

        # 满足终止条件则终止
        if should_stop:
            break

        # 翻到下一页
        current_pn = target.get_next_page_number(current_pn, g)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="anobbs-client",
    version="0.0.0",
    author="FToovvr",
    author_email="FToovvr@protonmail.com",
    description="Simple AnoBBS API wrapper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FToovvr/anobbs-api-wrapper-py",
    packages=setuptools.find_packages(exclude=("test",)),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
        "Natural Language :: Chinese (Simplified)",
    ],
    python_requires='>=3.8',
    install_requires=[
        "python-dateutil>=2.8.1",
        "requests>=2.24",
        "requests-toolbelt>=0.9.1",
    ],
)



from typing import OrderedDict, Any, List, Tuple, Optional
from dataclasses import dataclass

import json
import re
from datetime import datetime

from dateutil import tz


datetime_re = re.compile(r"^(.*?)\(.\)(.*?)$")
local_tz = tz.gettz("Asia/Shanghai")


@dataclass
class Post:
    """
    帖子基类。

    TODO: 回应比串多了一个目前用途未知的 ``status`` 字段，
          所以回应应该对应单独一个子类？
    """

    _raw: OrderedDict[str, Any]

    def __init__(self, data: OrderedDict[str, Any]):
        self._raw = data

    def raw_copy(self) -> OrderedDict[str, Any]:
        return OrderedDict(self._raw)

    @property
    def id(self) -> int:
        return int(self._raw["id"])

    @property
    def attachment_base(self) -> Optional[str]:
        return _none_if(self._raw["img"], "")

    @property
    def attachment_extension(self) -> Optional[str]:
        return _none_if(self._raw["ext"], "")

    @property
    def created_at_raw_text(self) -> str:
        return self._raw["now"]

    @property
    def created_at(self) -> datetime:
        g = datetime_re.match(self.created_at_raw_text)
        dt = datetime.strptime(f"{g[1]} {g[2]}", "%Y-%m-%d %H:%M:%S")
        return dt.replace(tzinfo=local_tz)

    @property
    def user_id(self) -> str:
        return self._raw["userid"]

    @property
    def name(self) -> Optional[str]:
        return _none_if(self._raw["name"], "无名氏")

    @property
    def email(self) -> Optional[str]:
        return _none_if(self._raw["email"], "")

    @property
    def title(self) -> Optional[str]:
        return _none_if(self._raw["title"], "无标题")

    @property
    def content(self) -> str:
        return self._raw["content"]

    @property
    def marked_sage(self) -> bool:
        return self._raw["sage"] != "0"

    @property
    def marked_admin(self) -> bool:
        return self._raw["admin"] != "0"

    def to_json(self) -> str:
        return json.dumps(self._raw, indent=2, ensure_ascii=False)


def _none_if(content, none_mark) -> Optional[Any]:
    if content == none_mark:
        return None
    return content


@dataclass
class ThreadBody(Post):
    """
    串基类。

    串在这里专指「一串帖子」中的头部，即「一楼」。
    """

    _total_reply_count: int

    def __init__(self, data: OrderedDict[str, Any], _total_reply_count: Optional[int] = None):
        super(ThreadBody, self).__init__(data)

        if _total_reply_count != None:
            self._total_reply_count = _total_reply_count
        else:
            self._total_reply_count = int(self._raw["replyCount"])
        # 不 pop 来保持顺序
        self._raw["replyCount"] = None

    def raw_copy(self, keeps_reply_count: bool = True, _keeps_replies_slot=False) -> OrderedDict[str, Any]:
        copy = super(ThreadBody, self).raw_copy()
        if keeps_reply_count:
            copy["replyCount"] = str(self._total_reply_count)
        else:
            copy.pop("replyCount")
        if not _keeps_replies_slot:
            copy.pop("replys", None)
        return copy

    @property
    def total_reply_count(self) -> int:
        return self._total_reply_count


@dataclass
class ThreadPage(ThreadBody):
    """
    是 ``get_thread_page`` 返回的串页面，包含其自身 body 及该页的各回复帖。
    """

    _replies: List[Post]

    def __init__(self, data: OrderedDict[str, Any]):
        super(ThreadPage, self).__init__(data)

        self._replies = list(map(lambda post: Post(post), self._raw["replys"]))
        # 不 pop 来保持顺序
        self._raw["replys"] = None

    def raw_copy(self) -> OrderedDict[str, Any]:
        copy = super(ThreadPage, self).raw_copy(_keeps_replies_slot=True)
        copy["replys"] = self.replies
        return copy

    @property
    def body(self) -> ThreadBody:
        return ThreadBody(self._raw, _total_reply_count=self._total_reply_count)

    @property
    def replies(self) -> List[Post]:
        return self._replies

    @replies.setter
    def replies(self, replies: List[Post]):
        self._replies = replies

    def to_json(self) -> str:
        data = self.raw_copy()
        if self._replies != None:
            data["replys"] = list(
                map(lambda post: post.raw_copy(), self._replies))
        else:
            data.pop("replys", None)

        return json.dumps(data, indent=2, ensure_ascii=False)


@dataclass
class BoardThread(ThreadPage):
    """
    组成 ``get_board_page`` 返回的各串的 body。
    """

    def __init__(self, data: OrderedDict[str, Any]):
        super(BoardThread, self).__init__(data)

    @property
    def last_modified_time(self) -> datetime:
        if len(self.replies) == 0:
            return self.created_at
        return self.replies[-1].created_at


@dataclass
class TimelineThread(BoardThread):
    """
    暂未使用，预计用于组成时间线。
    """

    def __init__(self, data: OrderedDict[str, Any]):
        super(TimelineThread, self).__init__(data)

    @property
    def board_id(self) -> int:
        return int(self._raw["fid"])


Board = List[BoardThread]



from typing import Optional, Tuple, OrderedDict, Dict, Any
from dataclasses import dataclass, field

from datetime import datetime

import anobbsclient

from .walktarget import WalkTargetInterface


@dataclass
class ReversalThreadWalkTargetState():
    last_page_min_id: Optional[int] = None


@dataclass(frozen=True)
class ReversalThreadWalkTarget(WalkTargetInterface):
    """
    从后向前，一路游荡到第一页，或者遇到的回复的串号/发布时间比预定的小/早时结束。

    可以保证不遗漏地归档串。
    """

    thread_id: int
    """要遍历的串的串号。"""

    gatekeeper_post_id: Optional[int]
    """
    不需要登录能看到的串中最大的串号。

    用于检测是否发生卡页。

    如果全程无需登录，无需设置；
    如果需要登录否则可能卡页，应设为「守门页」（一般为第100页）最后一串的串号。
    """

    # overriding
    start_page_number: int

    stop_before_post_id: Optional[int] = field(default=None)
    """
    串号的停止条件。
    在看到小于等于此串号的回复时停止。

    停止条件互相排斥，只可设定1处。
    如果没有停止条件，则会一路游荡到第一页。
    """

    expected_stop_page_number: Optional[int] = field(default=None)
    """
    如果串号是停止条件，可以设置此值，表示停止串号应该在此页或越过此页后找到。

    由于a岛只存在删串不存在增串，已知一个串号曾经出现在某页，
    则「反向遍历时先于预期遇到/超越此串号」只可能是出现「卡页」现象，
    服务器返回了这么遍历会更靠后才会遇到的页面。
    """

    stop_before_datetime: Optional[datetime] = field(default=None)
    """
    发布时间的停止条件。
    在看到早于此时间的回复时停止。
    """

    def __post_init__(self):
        # 确保停止条件只有1处
        stop_condition_count = 0
        if self.stop_before_post_id is not None:
            stop_condition_count += 1
        if self.stop_before_datetime is not None:
            stop_condition_count += 1
        if stop_condition_count > 1:
            raise ValueError()  # TODO: 更明确的异常

    # overriding
    def create_state(self) -> ReversalThreadWalkTargetState:
        return ReversalThreadWalkTargetState()

    # overriding
    def get_page(self, current_page_number: int,
                 client: anobbsclient.Client, options: anobbsclient.RequestOptions
                 ) -> Tuple[anobbsclient.ThreadPage, anobbsclient.BandwidthUsage]:
        """
        获取当前页数对应的页面内容并返回。

        Parameters
        ----------
        current_page_number : int
            当前页数。
        client : anobbsclient.Client
            用于发送请求的客户端。
        options : anobbsclient.RequestOptions
            要传入客户端的外部的请求设置。
        """
        return client.get_thread_page(
            id=self.thread_id, page=current_page_number,
            options=options, for_analysis=True,
        )

    # overriding
    def check_gatekept(self, current_page_number: int,
                       current_page: anobbsclient.ThreadPage,
                       client: anobbsclient.Client, options: anobbsclient.RequestOptions,
                       g: ReversalThreadWalkTargetState):
        """
        检查是否发生卡页。

        如检测的发生卡页，会直接相应的抛出异常。

        Parameters
        ----------
        current_page_number : int
            当前页数。
        current_page : anobbsclient.ThreadPage
            当前获取到的页。
        client : anobbsclient.Client
            用于发送请求的客户端。
        options : anobbsclient.RequestOptions
            要传入客户端的外部的请求设置。
        """

        # 通过检查本页与上一页是否一致来检测是否卡页。
        # 当前页应该至少有1串比上一页的所有串号要小。
        # 极端情况下，在获取两页期间，小于等于本页的地方有19串被删会导致误判，这里不考虑
        if g.last_page_min_id is not None \
                and current_page.replies[0].id >= g.last_page_min_id:
            raise anobbsclient.GatekeptException(
                context="previous_page_min_post_id",
                current_page_number=current_page_number,
                gatekeeper_post_id=g.last_page_min_id,
            )
        g.last_page_min_id = current_page.replies[0].id

        # 如果确认此页不登录会卡页，
        # 那通过对比「守门串号」（即不登录能看到的最后一串的串号）来检测是否卡页。
        # 少数情况下，遍历期间有19串被删导致「守门页」的串号要比起初获得的「守门串号」大，
        # 不过自从遍历第二页起就可以用前面对比两页的方法检测是否发生卡页，因此不是大问题
        if self.gatekeeper_post_id is not None \
                and client.thread_page_requires_login(current_page_number) \
                and current_page.replies[0].id <= self.gatekeeper_post_id:
            raise anobbsclient.GatekeptException(
                context="gatekeeper_post_id",
                current_page_number=current_page_number,
                gatekeeper_post_id=self.gatekeeper_post_id,
            )

        # 如果 expected_stop_page_number 超过守门页，就用 stop_before_post_id 判断是否卡页
        if self.expected_stop_page_number is not None \
                and current_page_number > self.expected_stop_page_number \
                and (True or client.thread_page_requires_login(self.expected_stop_page_number)) \
                and current_page.replies[0].id <= self.expected_stop_page_number:
            # 当前页面还没到预期停止页面，却得到了串号不大于停止串号的串，代表卡页了。
            # 当然，极端情况下，在页数小于等于预期停止页数的页面连删19串以上，
            # 会导致此验证法失效，这里不考虑
            raise anobbsclient.GatekeptException(
                context="lower_bound_post_id",
                current_page_number=current_page_number,
                gatekeeper_post_id=self.stop_before_post_id,
            )

    # overriding
    def should_stop(self, current_page: anobbsclient.ThreadPage, current_page_number: int,
                    client: anobbsclient.Client, options: anobbsclient.RequestOptions,
                    g: ReversalThreadWalkTargetState) -> bool:
        """
        返回是否已经达成停止条件。
        会在获取页面后被调用。

        TODO: 考虑将时间作为停止条件。

        Parameters
        ----------
        current_page : anobbsclient.ThreadPage
            当前获取到的页。
        next_page_number : int
            当前页数。
        """

        # 如果有设置停止串号，
        # 则试着找到停止串号（如果没有，则比它小且离它最近的串号）在回复中的 index
        if (self.stop_before_post_id or self.stop_before_datetime) is not None:
            stop_i = None
            for (i, reply) in enumerate(current_page.replies):
                if False \
                    or (self.stop_before_post_id is not None
                        and reply.id <= self.stop_before_post_id) \
                    or (self.stop_before_datetime is not None
                        and reply.created_at < self.stop_before_datetime):
                    stop_i = i
                else:
                    break
            if stop_i is not None:  # 找到了上述串号
                # 截掉更早的回复。
                # TODO: 允许自定义是否截掉
                current_page.replies = current_page.replies[stop_i+1:]
                return True

        if current_page_number == 1:
            # 越过第1页
            return True

    # overriding
    def get_next_page_number(self, current_page_number: int, g: ReversalThreadWalkTargetState):
        """
        获取将要获取的下一页的页数。

        Parameters
        ----------
        current_page_number : int
            当前页的页数。
        """
        return current_page_number - 1
